import pathlib

OVERWRITE_SHORT_OPTION = "--overwrite"
OVERWRITE_LONG_OPTION = "--overwrite-output-data"
OVERWRITE_OPTION_HELP_TEXT = (
    "If specified and if the file already exists on destination, then it will be "
    "overwritten instead of creating new one with unique index."
)

FORCE_DOWNLOAD_CLI_PROMPT_MESSAGE = "Do you want to proceed with download?"

DEFAULT_CLIENT_BASE_DIRECTORY = (
    pathlib.Path.home() / ".copernicus_marine_client"
)


def get_unique_filename(
    filepath: pathlib.Path, overwrite_option: bool
) -> pathlib.Path:
    if not overwrite_option:
        parent = filepath.parent
        filename = filepath.stem
        extension = filepath.suffix
        counter = 1

        while filepath.exists():
            filepath = parent / (
                filename + "_(" + str(counter) + ")" + extension
            )
            counter += 1

    return filepath



from dataclasses import dataclass
from datetime import datetime
from typing import Optional


@dataclass
class LatitudeParameters:
    minimal_latitude: Optional[float]
    maximal_latitude: Optional[float]


@dataclass
class LongitudeParameters:
    minimal_longitude: Optional[float]
    maximal_longitude: Optional[float]


@dataclass
class GeographicalParameters:
    latitude_parameters: LatitudeParameters
    longitude_parameters: LongitudeParameters


@dataclass
class TemporalParameters:
    start_datetime: Optional[datetime]
    end_datetime: Optional[datetime]


@dataclass
class DepthParameters:
    minimal_depth: Optional[float]
    maximal_depth: Optional[float]
    vertical_dimension_as_originally_produced: bool = True



from datetime import datetime
from decimal import Decimal
from typing import List, Literal, Optional, Union

import numpy as np
import xarray as xr

from copernicus_marine_client.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    LatitudeParameters,
    LongitudeParameters,
    TemporalParameters,
)

COORDINATES_LABEL = {
    "latitude": ["latitude", "nav_lat", "x"],
    "longitude": ["longitude", "nav_lon", "y"],
    "time": ["time_counter", "time"],
    "depth": ["depth", "deptht", "elevation"],
}


class MinimalLongitudeGreaterThanMaximalLongitude(Exception):
    pass


def dataset_custom_sel(
    dataset: xr.Dataset,
    coord_type: Literal["latitude", "longitude", "depth", "time"],
    coord_selection: Union[float, slice, datetime, None],
    method: Union[str, None] = None,
) -> xr.Dataset:
    for coord_label in COORDINATES_LABEL[coord_type]:
        if coord_label in dataset.coords:
            dataset = dataset.sel(
                {coord_label: coord_selection}, method=method
            )
    return dataset


def _latitude_subset(
    dataset: xr.Dataset, latitude_parameters: LatitudeParameters
) -> xr.Dataset:
    minimal_latitude = latitude_parameters.minimal_latitude
    maximal_latitude = latitude_parameters.maximal_latitude
    if minimal_latitude is not None or maximal_latitude is not None:
        latitude_selection = (
            minimal_latitude
            if minimal_latitude == maximal_latitude
            else slice(minimal_latitude, maximal_latitude)
        )
        latitude_method = (
            "nearest" if minimal_latitude == maximal_latitude else None
        )
        dataset = dataset_custom_sel(
            dataset, "latitude", latitude_selection, latitude_method
        )
    return dataset


def _longitude_subset(
    dataset: xr.Dataset, longitude_parameters: LongitudeParameters
) -> xr.Dataset:
    def _update_dataset_attributes(dataset: xr.Dataset):
        for coord_label in COORDINATES_LABEL["longitude"]:
            if coord_label in dataset.coords:
                attrs = dataset[coord_label].attrs
                if "valid_min" in attrs:
                    attrs["valid_min"] += 180
                if "valid_max" in attrs:
                    attrs["valid_max"] += 180
                dataset = dataset.assign_coords(
                    {coord_label: dataset[coord_label] % 360}
                ).sortby(coord_label)
                dataset[coord_label].attrs = attrs
        return dataset

    minimal_longitude = longitude_parameters.minimal_longitude
    maximal_longitude = longitude_parameters.maximal_longitude
    if minimal_longitude is not None or maximal_longitude is not None:
        if minimal_longitude is not None and maximal_longitude is not None:
            if minimal_longitude > maximal_longitude:
                raise MinimalLongitudeGreaterThanMaximalLongitude(
                    "--minimal-longitude option must be smaller "
                    "or equal to --maximal-longitude"
                )
            if maximal_longitude - minimal_longitude >= 360:
                longitude_selection: Union[float, slice, None] = None
            elif minimal_longitude == maximal_longitude:
                longitude_selection = longitude_modulus(minimal_longitude)
                longitude_method = "nearest"
            else:
                minimal_longitude_modulus = longitude_modulus(
                    minimal_longitude
                )
                maximal_longitude_modulus = longitude_modulus(
                    maximal_longitude
                )
                if maximal_longitude_modulus < minimal_longitude_modulus:
                    maximal_longitude_modulus += 360
                    dataset = _update_dataset_attributes(dataset)
                longitude_selection = slice(
                    minimal_longitude_modulus,
                    maximal_longitude_modulus,
                )
                longitude_method = None
        else:
            longitude_selection = slice(minimal_longitude, maximal_longitude)
            longitude_method = None

        if longitude_selection is not None:
            dataset = dataset_custom_sel(
                dataset, "longitude", longitude_selection, longitude_method
            )
    return dataset


def _temporal_subset(
    dataset: xr.Dataset, temporal_parameters: TemporalParameters
) -> xr.Dataset:
    start_datetime = temporal_parameters.start_datetime
    end_datetime = temporal_parameters.end_datetime
    if start_datetime is not None or end_datetime is not None:
        temporal_selection = (
            start_datetime
            if start_datetime == end_datetime
            else slice(start_datetime, end_datetime)
        )
        temporal_method = "nearest" if start_datetime == end_datetime else None
        dataset = dataset_custom_sel(
            dataset, "time", temporal_selection, temporal_method
        )
    return dataset


def _depth_subset(
    dataset: xr.Dataset, depth_parameters: DepthParameters
) -> xr.Dataset:
    def convert_elevation_to_depth(dataset: xr.Dataset):
        if "elevation" in dataset.dims:
            attrs = dataset["elevation"].attrs
            dataset = dataset.reindex(elevation=dataset.elevation[::-1])
            dataset["elevation"] = dataset.elevation * (-1)
            dataset = dataset.rename({"elevation": "depth"})
            dataset.depth.attrs = attrs
        return dataset

    minimal_depth = depth_parameters.minimal_depth
    maximal_depth = depth_parameters.maximal_depth
    if minimal_depth is not None or maximal_depth is not None:
        if "elevation" in dataset.dims:
            minimal_depth = (
                minimal_depth * -1.0 if minimal_depth is not None else None
            )
            maximal_depth = (
                maximal_depth * -1.0 if maximal_depth is not None else None
            )
            minimal_depth, maximal_depth = maximal_depth, minimal_depth

        depth_selection = (
            minimal_depth
            if minimal_depth == maximal_depth
            else slice(minimal_depth, maximal_depth)
        )
        depth_method = "nearest" if minimal_depth == maximal_depth else None
        dataset = dataset_custom_sel(
            dataset, "depth", depth_selection, depth_method
        )
    if depth_parameters.vertical_dimension_as_originally_produced:
        dataset = convert_elevation_to_depth(dataset)
    return dataset


def subset(
    dataset,
    variables: Optional[List[str]],
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
) -> xr.Dataset:

    if variables:
        dataset = dataset[np.array(variables)]

    dataset = _latitude_subset(
        dataset, geographical_parameters.latitude_parameters
    )
    dataset = _longitude_subset(
        dataset, geographical_parameters.longitude_parameters
    )

    dataset = _temporal_subset(dataset, temporal_parameters)

    dataset = _depth_subset(dataset, depth_parameters)

    return dataset


def longitude_modulus(longitude: float) -> float:
    """
    Returns the equivalent longitude between -180 and 180
    """
    # We are using Decimal to avoid issue with rounding
    modulus = float(Decimal(str(longitude + 180)) % 360)
    # Modulus with python return a negative value if the denominator is negative
    # To counteract that, we add 360 if the result is < 0
    modulus = modulus if modulus >= 0 else modulus + 360
    return modulus - 180



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['copernicus_marine_client',
 'copernicus_marine_client.catalogue_parser',
 'copernicus_marine_client.command_line_interface',
 'copernicus_marine_client.download_functions']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3',
 'boto3>=1.28.2,<2.0.0',
 'cachier>=2.0.2',
 'click>=8.0.4',
 'dask>=2022.1.1',
 'motuclient==1.8.4',
 'netCDF4>=1.6.3',
 'numpy>=1.0',
 'owslib>=0.27.2',
 'pydap>=3.2.2',
 'requests>=2.27.1',
 'setuptools>=62.0.0',
 'tqdm>=4.65.0',
 'xarray>=2023.3.0',
 'zarr>=2.13.3']

entry_points = \
{'console_scripts': ['copernicus-marine = '
                     'copernicus_marine_client.command_line_interface.copernicus_marine:command_line_interface']}

setup_kwargs = {
    'name': 'copernicus-marine-client',
    'version': '0.8.13',
    'description': '',
    'long_description': '# Copernicus Marine Service client\n\nA library to facilitate the access of Copernicus Marine Service products and datasets.\n\n## Introduction\n\nThis package allows to recover products and datasets information from Command Line Interface,\nas well as download subsets and originally produced files.\n\n## Command Line Interface (CLI)\n\n### Command *describe*\n\nRetrieve information about all products as JSON:\n\n```txt\n> copernicus-marine describe\n{\n  "products": [\n    {\n      "title": "Antarctic Sea Ice Extent from Reanalysis",\n      "product_id": "ANTARCTIC_OMI_SI_extent",\n      "thumbnail_url": "https://catalogue.marine.copernicus.eu/documents/IMG/ANTARCTIC_OMI_SI_extent.png",\n      "production_center": "Mercator Oc\\u00e9an International",\n      "creation_datetime": "2018-02-12",\n      "modified_datetime": "2018-02-12",\n    }\n    ...\n  ]\n}\n```\n\nRetrieve all information about datasets as JSON:\n\n```txt\n> copernicus-marine describe --include-datasets\n{\n  "products": [\n    {\n      "title": "Antarctic Sea Ice Extent from Reanalysis",\n      "product_id": "ANTARCTIC_OMI_SI_extent",\n      "thumbnail_url": "https://catalogue.marine.copernicus.eu/documents/IMG/ANTARCTIC_OMI_SI_extent.png",\n      "production_center": "Mercator Oc\\u00e9an International",\n      "creation_datetime": "2018-02-12",\n      "modified_datetime": "2018-02-12",\n      "datasets": [\n        {\n          "dataset_id": "antarctic_omi_si_extent",\n          "dataset_name": "antarctic_omi_si_extent",\n          "services": [\n            {\n              "protocol": "ftp",\n              "uri": "ftp://my.cmems-du.eu/Core/ANTARCTIC_OMI_SI_extent/antarctic_omi_si_extent"\n            }\n          ],\n          "variables": []\n        }\n      ]\n    },\n    ...\n  ]\n}\n```\n\n### Command *login*\n\nCreate the configuration files for access to the copernicus marine service:\n\'.dodsrc\', \'.netrc\', \'.motuclient-python.ini\'.\nThe directory to store these configuration files can be modified by the user using the "config-file-directory" option\nbut beware as it should also be passed to the *subset* and *get* command afterwards.\nBy default, if the configuration files already exist, the user is asked for confirmation to overwrite them.\n\nExample:\n```\n> copernicus marine login\n< Username :\n< Password :\n> INFO     - root - Configuration files stored in ${HOME}\\.copernicus_marine_client\n```\n\n### Command *subset*\n\nDownload a dataset subset, based on dataset id, variable names and attributes slices:\n\n```txt\n> copernicus-marine subset -i METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2 -v analysed_sst -v sea_ice_fraction -t 2021-01-01 - T 2021-01-03 -x 0.0 -X 0.1 -y 0.0 -Y 0.1\n\n< Username:\n< Password:\n< Trying to download as one file...\n```\n\nFile downloaded to ./{dataset_id}.{nc/zarr} if not specified otherwise (through -o/--output-directory and -f/--output-filename options). If the output-filename argument ends with \'.nc\' suffix, the file will be downloaded as a netCDF file.\n\nOptions `--minimal-longitude FLOAT` and `--maximal-longitude FLOAT` work as follow:\n* `--minimal-longitude` must be lower or equal to `--maximal-longitude`\n* If `maximal-longitude - minimal-longitude >= 360`, return the full dataset.\n* Else:\n    - If the requested range **does not cross** the antemeridian, the result dataset will be between -180° and 180°.\n    - If it **does cross** the antemeridian, the result dataset will be between 0° and 360°.\n\nNote that you can request any longitudes you want. A modulus is applied to bring the result between -180° and 360°. For example, if you request [530°, 560°], the result dataset will be in [170°, 190°].\n\n\nThe `--vertical-dimension-as-originally-produced BOOLEAN` option is used to obtain a dataset whose vertical dimension (the z-axis) is as in the originally produced dataset. This dimension will be named `depth` with descending positive values. Its default value is `true`.\n\n### Command *get*\n\nDownload an originally produced file (or files), based on dataset id or path to files:\n\nExample:\n\n```txt\n> copernicus-marine get -u ftp://my.cmems-du.eu/Core/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/\n\n< Username:\n< Password:\n< You requested the download of the following files:\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202207.nc - 3.27 MB\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202208.nc - 3.29 MB\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202209.nc - 3.28 MB\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202210.nc - 3.26 MB\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202211.nc - 3.26 MB\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202212.nc - 3.26 MB\n\nTotal size of the download: 19.62 MB\n\n\nDo you want to proceed with download? [Y/n]:\n```\n\nFile(s) downloaded to ./{path}/{filename} if not specified otherwise:\n\n- "--output-path" specifies a directory to dump the files in\n- "--no-directories" to not recreate the folder structure\n\nIf not specified otherwise, after the header display with a summary of the request,\nthe user is asked for confirmation:\n\n- "--no-confirmation" to turn down the confirmation prompt\n- "--show-outputnames" to display the full paths of the outputs files\n\nOption `--filter TEXT` allows to specify a Unix shell-style wildcard pattern (see [fnmatch — Unix filename pattern matching](https://docs.python.org/3/library/fnmatch.html)) and select specific files. It work with both `--dataset-id` and `--dataset-url` options.\n\nOption `--regex TEXT` allows to specify a regular expression and select specific files. It work with both `--dataset-id` and `--dataset-url` options.\n\nExample:\n```\n> copernicus-marine get -u ftp://my.cmems-du.eu/Core/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m --regex ".*2022(08|09|10).nc"\nPassword:\nYou requested the download of the following files:\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202208.nc - 3.29 MB\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202209.nc - 3.28 MB\nCore/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m/2022/metoffice_foam1_amm7_NWS_DIATO_CPWC_mm202210.nc - 3.26 MB\n\nTotal size of the download: 9.82 MB\n\n\nDo you want to proceed with download? [Y/n]:\n```\n\n### The overwrite option\n\nBoth `get` and `subset` commands provide an `--overwrite-ouput-data` option.\nWhen not provided (default behavior), once the download has been accepted (or if the `--force-download` option was provided), if the file already exists on destination, then a new one with a unique index will be created.\nOn the other hand, if the `--overwrite-ouput-data` option is provided and the file already exists, then it\'ll be overwritten.\n\n### The `--help` argument\n\nIn any case, please remember that you can call the `--help` argument on any CLI option. This may save you some time or find what you need.\n\n## Installation\n\nUsing pip, for example:\n\n```shell\npip install copernicus-marine-client\n```\n\n## Technical details\n\nThis module is organized around two capabilities:\n\n- a catalogue, parsed from web requests, that contains informations on the available datasets\n- a downloader, to simplify the download of dataset files or subsets\n\nThe catalogue can be displayed by the user and is used by the downloader to link the user\nrequests with files or subset of files to retrieve.\nThe downloader will help the user download the needed datasets.\n\nA rigid format, specified in "request_structure.py" is used to ensure conformity of the information passed between the CLI command and the python functions.\n\nFor subset command, the format is:\n\n```python\n@dataclass\nclass SubsetRequest:\n    dataset_url: Optional[str] = None\n    dataset_id: Optional[str] = None\n    variables: Optional[List[str]] = None\n    minimal_longitude: Optional[float] = None\n    maximal_longitude: Optional[float] = None\n    minimal_latitude: Optional[float] = None\n    maximal_latitude: Optional[float] = None\n    minimal_depth: Optional[float] = None\n    maximal_depth: Optional[float] = None\n    start_datetime: Optional[datetime] = None\n    end_datetime: Optional[datetime] = None\n    output_directory: Optional[str] = None\n    output_filename: Optional[str] = None\n    force_download: Optional[bool] = None\n    force_protocol: Optional[str] = None\n```\n',
    'author': 'Copernicus Marine User Support',
    'author_email': 'servicedesk.cmems@mercator-ocean.eu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9',
}


setup(**setup_kwargs)



import asyncio
import logging
import logging.config
from copy import deepcopy
from dataclasses import dataclass
from itertools import groupby
from json import loads
from multiprocessing import Pool
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

import aiohttp
import requests
from cachier import cachier

from copernicus_marine_client.utils import DEFAULT_CLIENT_BASE_DIRECTORY


@dataclass
class CopernicusMarineDatasetService:
    protocol: str
    uri: str


@dataclass
class CopernicusMarineDatasetCoordinates:
    coordinates_id: str
    units: str
    minimum_value: float
    maximum_value: float
    step: Optional[float]
    values: Optional[str]


@dataclass
class CopernicusMarineDatasetVariable:
    short_name: str
    standard_name: str
    units: str
    bbox: Tuple[float, float, float, float]
    coordinates: list[CopernicusMarineDatasetCoordinates]


@dataclass
class CopernicusMarineProductDataset:
    dataset_id: str
    dataset_name: str
    services: list[CopernicusMarineDatasetService]
    variables: list[CopernicusMarineDatasetVariable]

    def get_available_protocols(self) -> list[str]:
        return list(map(lambda service: service.protocol, self.services))


@dataclass
class CopernicusMarineProductProvider:
    name: str
    roles: list[str]
    url: str
    email: str


@dataclass
class CopernicusMarineProduct:
    title: str
    product_id: str
    thumbnail_url: str
    description: str
    production_center: str
    creation_datetime: str
    modified_datetime: Optional[str]
    keywords: dict[str, str]
    datasets: list[CopernicusMarineProductDataset]


@dataclass
class CopernicusMarineCatalogue:
    products: list[CopernicusMarineProduct]

    def filter(self, tokens: list[str]):
        return filter_catalogue_with_strings(self, tokens)


OPENDAP_KEY = "opendap"
MOTU_KEY = "motu"
FTP_KEY = "ftp"
GEOCHUNKED_KEY = "geoChunked"
TIMECHUNKED_KEY = "timeChunked"
S3NATIVE_KEY = "native"
DOWNSAMPLED4_KEY = "downsampled4"

_S = TypeVar("_S")
_T = TypeVar("_T")


def map_parallel(
    function: Callable[[_S], _T], iterable: Iterable[_S]
) -> list[_T]:
    parallel_processes = 20
    with Pool(parallel_processes) as pool:
        return pool.map(function, iterable)


def map_reject_none(
    function: Callable[[_S], Optional[_T]], iterable: Iterable[_S]
) -> Iterable[_T]:
    return (element for element in map(function, iterable) if element)


def parse_catalogue(
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
) -> CopernicusMarineCatalogue:
    return parse_dissemination_unit_catalogue(
        overwrite_metadata_cache=overwrite_metadata_cache,
        no_metadata_cache=no_metadata_cache,
    )


async def _async_fetch_raw_product(session, product_id: str):
    async with session.get(product_url(product_id)) as resp:
        text_response = await resp.text()
        return loads(text_response)


async def _async_fetch_raw_products(product_ids: List[str]):
    async with aiohttp.ClientSession() as session:
        tasks = []
        for product_id in product_ids:
            tasks.append(
                asyncio.ensure_future(
                    _async_fetch_raw_product(session, product_id)
                )
            )

        return await asyncio.gather(*tasks)


@cachier(cache_dir=DEFAULT_CLIENT_BASE_DIRECTORY)
def _fetch_raw_products() -> list[dict[str, Any]]:
    response = requests.post(
        "https://data-be-prd.marine.copernicus.eu/api/datasets",
        json={"size": 1000, "includeOmis": True},
    )
    assert response.ok, response.text
    raw_catalogue: dict[str, Any] = loads(response.text)

    results = asyncio.run(
        _async_fetch_raw_products(raw_catalogue["datasets"].keys())
    )
    return results


def product_url(product_id: str) -> str:
    return (
        f"https://data-be-prd.marine.copernicus.eu/api/dataset/{product_id}"
        + "?variant=detailed-v2"
    )


def variable_title_to_standard_name(variable_title: str) -> str:
    return variable_title.lower().replace(" ", "_")


def variable_to_pick(layer: dict[str, Any]) -> bool:
    return (
        layer["variableId"] != "__DEFAULT__"
        and layer["subsetVariableIds"]
        and len(layer["subsetVariableIds"]) == 1
    )


def stac_services_per_dataset_id(
    stac_items: Dict[str, Dict]
) -> Dict[str, List[CopernicusMarineDatasetService]]:
    results: Dict[str, List[CopernicusMarineDatasetService]] = {}
    for stac_item in stac_items.values():
        stac_dataset_id = stac_item["id"]
        stac_assets = stac_item["assets"]
        for protocol, values in stac_assets.items():
            protocol_uri = values["href"]
            result = CopernicusMarineDatasetService(
                protocol=protocol, uri=protocol_uri
            )
            if stac_dataset_id in results:
                results[stac_dataset_id].append(result)
            else:
                results[stac_dataset_id] = [result]
    return results


def to_datasets(
    raw_services: dict[str, dict[str, str]],
    layers: dict[str, dict[str, Any]],
    stac_items: dict,
) -> list[CopernicusMarineProductDataset]:
    def to_service(
        protocol_uri: Tuple[str, str]
    ) -> CopernicusMarineDatasetService:
        return CopernicusMarineDatasetService(
            protocol=protocol_uri[0], uri=protocol_uri[1]
        )

    def to_variable(layer: dict[str, Any]) -> CopernicusMarineDatasetVariable:
        def to_coordinates(
            subset_attributes: Tuple[str, dict[str, Any]]
        ) -> CopernicusMarineDatasetCoordinates:
            coordinate_name = subset_attributes[0]
            values: Optional[str]
            if coordinate_name == "depth":
                values = layer.get("zValues")
            elif coordinate_name == "time":
                values = layer.get("tValues")
            else:
                values = None
            return CopernicusMarineDatasetCoordinates(
                coordinates_id=subset_attributes[0],
                units=subset_attributes[1]["units"],
                minimum_value=subset_attributes[1]["min"],
                maximum_value=subset_attributes[1]["max"],
                step=subset_attributes[1].get("step"),
                values=values,
            )

        return CopernicusMarineDatasetVariable(
            short_name=layer["variableId"],
            standard_name=variable_title_to_standard_name(
                layer["variableTitle"]
            ),
            units=layer["units"],
            bbox=layer["bbox"],
            coordinates=list(
                map(to_coordinates, layer["subsetAttrs"].items())
            ),
        )

    def to_service_stac(asset) -> Optional[CopernicusMarineDatasetService]:
        protocol = asset[0]
        protocol_uri = asset[1]["href"]
        if protocol != "omi":
            return CopernicusMarineDatasetService(
                protocol=protocol, uri=protocol_uri
            )
        else:
            return None

    @dataclass
    class DistinctDataset:
        dataset_id: str
        layer_elements: Iterable
        raw_services: Dict
        stac_items_values: Optional[Dict]

    def to_dataset(
        distinct_dataset: DistinctDataset,
    ) -> CopernicusMarineProductDataset:
        dataset_id = distinct_dataset.dataset_id
        layer_elements = list(distinct_dataset.layer_elements)
        services_portal = list(
            map(to_service, distinct_dataset.raw_services.items())
        )
        services_stac = (
            list(
                filter(
                    lambda x: x is not None,
                    map(
                        to_service_stac,
                        distinct_dataset.stac_items_values["assets"].items(),
                    ),
                )
            )
            if distinct_dataset.stac_items_values
            else []
        )

        services = services_portal + cast(
            List[CopernicusMarineDatasetService], services_stac
        )

        return CopernicusMarineProductDataset(
            dataset_id=dataset_id,
            dataset_name=layer_elements[0]["subdatasetTitle"],
            services=services,
            variables=list(
                map(to_variable, filter(variable_to_pick, layer_elements))
            ),
        )

    def construct_unique_dataset(group_layer) -> DistinctDataset:
        dataset_id_from_layer = group_layer[0]
        dataset_layer_elements = group_layer[1]
        dataset_raw_services = raw_services[dataset_id_from_layer]

        for stac_dataset_id, stac_items_values in stac_items.items():
            if stac_dataset_id.startswith(dataset_id_from_layer):
                if (
                    "--ext--"
                    in stac_dataset_id.split(dataset_id_from_layer)[-1]
                ):
                    continue
                else:
                    return DistinctDataset(
                        dataset_id=dataset_id_from_layer,
                        layer_elements=dataset_layer_elements,
                        raw_services=dataset_raw_services,
                        stac_items_values=stac_items_values,
                    )
        else:
            return DistinctDataset(
                dataset_id=dataset_id_from_layer,
                layer_elements=dataset_layer_elements,
                raw_services=dataset_raw_services,
                stac_items_values=None,
            )

    groups_layers = groupby(
        layers.values(), key=lambda layer: layer["subdatasetId"]
    )
    distinct_datasets = map(construct_unique_dataset, groups_layers)

    return sorted(
        map(to_dataset, distinct_datasets),
        key=lambda distinct_dataset: distinct_dataset.dataset_id,
    )


def _parse_product(raw_product: dict[str, Any]) -> CopernicusMarineProduct:
    return CopernicusMarineProduct(
        title=raw_product["title"],
        product_id=raw_product["id"],
        thumbnail_url=raw_product["thumbnailUrl"],
        description=raw_product["abstract"],
        production_center=raw_product["originatingCenter"],
        creation_datetime=raw_product["creationDate"],
        modified_datetime=raw_product.get("modifiedDate"),
        keywords=raw_product["keywords"],
        datasets=to_datasets(
            raw_product["services"],
            raw_product["layers"],
            raw_product["stacItems"],
        ),
    )


def parse_dissemination_unit_catalogue(
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
) -> CopernicusMarineCatalogue:
    raw_products: list[dict[str, Any]] = _fetch_raw_products(
        overwrite_cache=overwrite_metadata_cache,
        ignore_cache=no_metadata_cache,
    )

    return CopernicusMarineCatalogue(
        products=sorted(
            map(_parse_product, raw_products),
            key=lambda product: product.product_id,
        ),
    )


# ---------------------------------------
# --- Utils function on any catalogue ---
# ---------------------------------------


class ProtocolNotAvailable(Exception):
    ...


def protocol_not_available_error(
    dataset_id: str, protocols: list[str]
) -> ProtocolNotAvailable:
    return ProtocolNotAvailable(
        f"Available protocols for dataset {dataset_id}: {protocols}"
    )


def get_protocol_url_from_id(
    dataset_id: str,
    protocol_key_order: list[str],
    overwrite_metadata_cache,
    no_metadata_cache,
) -> Tuple[str, str]:
    catalogue = parse_catalogue(overwrite_metadata_cache, no_metadata_cache)
    dataset_urls = (
        get_dataset_url_from_id(catalogue, dataset_id, protocol)
        for protocol in protocol_key_order
    )
    try:
        dataset_url = next(filter(lambda dataset: dataset, dataset_urls))
        protocol = get_protocol_from_url(dataset_url)
    except StopIteration as exception:
        raise protocol_not_available_error(
            dataset_id, protocol_key_order
        ) from exception
    return protocol, dataset_url


def get_dataset_from_id(
    catalogue: CopernicusMarineCatalogue, dataset_id: str
) -> CopernicusMarineProductDataset:
    for product in catalogue.products:
        for dataset in product.datasets:
            if dataset_id == dataset.dataset_id:
                return dataset
    error = KeyError(
        f"The requested dataset '{dataset_id}' was not found in the catalogue,"
        " you can use 'copernicus-marine describe --include-datasets "
        "-c <search_token>' to find the dataset id"
    )
    logging.error(error)
    raise error


def get_product_from_url(
    catalogue: CopernicusMarineCatalogue, dataset_url: str
) -> CopernicusMarineProduct:
    """
    Return the product object, with its dataset list filtered
    """
    filtered_catalogue = filter_catalogue_with_strings(
        catalogue, [dataset_url]
    )
    if filtered_catalogue is None:
        error = TypeError("filtered catalogue is empty")
        raise error
    return filtered_catalogue.products[0]


def get_protocol_from_url(dataset_url) -> str:
    if dataset_url.startswith("ftp://"):
        protocol = FTP_KEY
    elif "/motu-web/Motu" in dataset_url:
        protocol = MOTU_KEY
    elif "/wms/" in dataset_url:
        protocol = "OGC:WMS:getCapabilities"
    elif "/thredds/dodsC/" in dataset_url:
        protocol = OPENDAP_KEY
    elif "/mdl-arco-time/" in dataset_url:
        protocol = TIMECHUNKED_KEY
    elif "/mdl-arco-geo/" in dataset_url:
        protocol = GEOCHUNKED_KEY
    elif "/mdl-native/" in dataset_url:
        protocol = S3NATIVE_KEY
    else:
        exception = ValueError(f"No protocol matching url: {dataset_url}")
        logging.error(exception)
        raise exception
    return protocol


def get_service_url(
    dataset: CopernicusMarineProductDataset, protocol: str
) -> str:
    service_urls = iter(
        [
            service.uri
            for service in dataset.services
            if service.protocol == protocol
        ]
    )
    return next(service_urls, "")


def get_dataset_url_from_id(
    catalogue: CopernicusMarineCatalogue, dataset_id: str, protocol: str
) -> str:
    dataset = get_dataset_from_id(catalogue, dataset_id)
    return get_service_url(dataset, protocol)


def filter_catalogue_with_strings(
    catalogue: CopernicusMarineCatalogue, tokens: list[str]
) -> Optional[CopernicusMarineCatalogue]:
    filtered_catalogue = deepcopy(catalogue)
    return find_match_object(filtered_catalogue, tokens)


def find_match_object(value: Any, tokens: list[str]) -> Any:
    match: Any
    if isinstance(value, str):
        match = find_match_string(value, tokens)
    elif isinstance(value, list):
        match = find_match_list(value, tokens)
    elif hasattr(value, "__dict__"):
        match = find_match_dict(value, tokens)
    else:
        match = None
    return match


def find_match_string(string: str, tokens: list[str]) -> Optional[str]:
    return string if any(token in string for token in tokens) else None


def find_match_list(object_list: list[Any], tokens) -> Optional[list[Any]]:
    def find_match(element: Any) -> Optional[Any]:
        return find_match_object(element, tokens)

    filtered_list: list[Any] = list(map_reject_none(find_match, object_list))
    return filtered_list if filtered_list else None


def find_match_dict(
    structure: dict[str, Any], tokens
) -> Optional[dict[str, Any]]:
    filtered_dict = {
        key: find_match_object(value, tokens)
        for key, value in structure.__dict__.items()
        if find_match_object(value, tokens)
    }

    found_match = any(filtered_dict.values())
    if found_match:
        new_dict = dict(structure.__dict__, **filtered_dict)
        structure.__dict__ = new_dict
    return structure if found_match else None
